@@ -59,6 +59,31 @@ def parse_labeled_data(df):
     return labeled_data
 
 
+def parse_feed_df(df: DataFrame, include_id):
+    """
+    Convert a df into batch format for feeding
+
+    :param df: DataFrame with the following required columns ["id"]. Additional columns are assumed to be fields.
+    :param include_id: Include id on the fields to be fed.
+    :return: List of Dict containing 'id' and 'fields'.
+    """
+    required_columns = ["id"]
+    assert all(
+        [x in list(df.columns) for x in required_columns]
+    ), "DataFrame needs at least the following columns: {}".format(required_columns)
+    records = df.to_dict(orient="records")
+    batch = [
+        {
+            "id": record["id"],
+            "fields": record
+            if include_id
+            else {k: v for k, v in record.items() if k not in ["id"]},
+        }
+        for record in records
+    ]
+    return batch
+
+
 class Vespa(object):
     def __init__(
         self,
@@ -66,6 +91,7 @@ class Vespa(object):
         port: Optional[int] = None,
         deployment_message: Optional[List[str]] = None,
         cert: Optional[str] = None,
+        key: Optional[str] = None,
         output_file: IO = sys.stdout,
         application_package: Optional[ApplicationPackage] = None,
     ) -> None:
@@ -75,7 +101,9 @@ class Vespa(object):
         :param url: Vespa instance URL.
         :param port: Vespa instance port.
         :param deployment_message: Message returned by Vespa engine after deployment. Used internally by deploy methods.
-        :param cert: Path to certificate and key file.
+        :param cert: Path to certificate and key file in case the 'key' parameter is none. If 'key' is not None, this
+            should be the path of the certificate file.
+        :param key: Path to the key file.
         :param output_file: Output file to write output messages.
         :param application_package: Application package definition used to deploy the application.
 
@@ -92,6 +120,7 @@ class Vespa(object):
         self.port = port
         self.deployment_message = deployment_message
         self.cert = cert
+        self.key = key
         self._application_package = application_package
 
         if port is None:
@@ -140,6 +169,26 @@ class Vespa(object):
         else:
             return "Vespa({})".format(self.url)
 
+    def _infer_schema_name(self):
+        if not self._application_package:
+            raise ValueError(
+                "Application Package not available. Not possible to infer schema name."
+            )
+
+        try:
+            schema = self._application_package.schema
+        except AssertionError:
+            raise ValueError(
+                "Application has more than one schema. Not possible to infer schema name."
+            )
+
+        if not schema:
+            raise ValueError(
+                "Application has no schema. Not possible to infer schema name."
+            )
+
+        return schema.name
+
     def get_application_status(self) -> Optional[Response]:
         """
         Get application status.
@@ -149,22 +198,22 @@ class Vespa(object):
         with VespaSync(self) as sync_app:
             return sync_app.get_application_status()
 
-    def get_model_endpoint(
-        self, model_name: Optional[str] = None
-    ) -> Optional[Response]:
+    def get_model_endpoint(self, model_id: Optional[str] = None) -> Optional[Response]:
         """Get model evaluation endpoints."""
 
         with VespaSync(self) as sync_app:
-            return sync_app.get_model_endpoint(model_name=model_name)
+            return sync_app.get_model_endpoint(model_id=model_id)
 
     def _build_query_body(
         self,
         query: Optional[str] = None,
         query_model: Optional[QueryModel] = None,
         recall: Optional[Tuple] = None,
-        **kwargs
+        **kwargs,
     ) -> Dict:
         assert query is not None, "No 'query' specified."
+        if not query_model:
+            query_model = self.get_default_query_model()
         assert query_model is not None, "No 'query_model' specified."
         body = query_model.create_body(query=query)
         if recall is not None:
@@ -187,7 +236,7 @@ class Vespa(object):
         query_model: Optional[QueryModel] = None,
         debug_request: bool = False,
         recall: Optional[Tuple] = None,
-        **kwargs
+        **kwargs,
     ) -> VespaQueryResponse:
         """
         Send a query request to the Vespa application.
@@ -210,7 +259,116 @@ class Vespa(object):
                 query_model=query_model,
                 debug_request=debug_request,
                 recall=recall,
-                **kwargs
+                **kwargs,
+            )
+
+    def _query_batch_sync(
+        self,
+        body_batch: Optional[List[Dict]],
+        query_batch: Optional[List[str]],
+        query_model: Optional[QueryModel],
+        recall: Optional[List[Tuple]],
+        **kwargs,
+    ):
+        if body_batch:
+            return [self.query(body=body, **kwargs) for body in body_batch]
+        else:
+            if recall:
+                return [
+                    self.query(query=q, query_model=query_model, recall=r, **kwargs)
+                    for (q, r) in zip(query_batch, recall)
+                ]
+            else:
+                return [
+                    self.query(query=query, query_model=query_model, **kwargs)
+                    for query in query_batch
+                ]
+
+    async def _query_batch_async(
+        self,
+        body_batch: Optional[List[Dict]],
+        query_batch: Optional[List[str]],
+        query_model: Optional[QueryModel],
+        recall: Optional[List[Tuple]],
+        connections,
+        total_timeout,
+        **kwargs,
+    ):
+        async with VespaAsync(
+            app=self, connections=connections, total_timeout=total_timeout
+        ) as async_app:
+            return await async_app.query_batch(
+                body_batch=body_batch,
+                query_batch=query_batch,
+                query_model=query_model,
+                recall=recall,
+                **kwargs,
+            )
+
+    def query_batch(
+        self,
+        body_batch: Optional[List[Dict]] = None,
+        query_batch: Optional[List[str]] = None,
+        query_model: Optional[QueryModel] = None,
+        recall: Optional[List[Tuple]] = None,
+        asynchronous=True,
+        connections: Optional[int] = 100,
+        total_timeout: int = 100,
+        **kwargs,
+    ):
+        """
+        Send queries in batch to a Vespa app.
+
+        :param body_batch: A list of dict containing all the request parameters. Set to None if using 'query_batch'.
+        :param query_batch: A list of query strings. Set to None if using 'body_batch'.
+        :param query_model: Query model to use when sending query strings. Set to None if using 'body_batch'.
+        :param recall: List of tuples, one for each query. Tuple of size 2 where the first element is the name
+            of the field to use to recall and the second element is a list of the values to be recalled.
+        :param asynchronous: Set True to send data in async mode. Default to True.
+        :param connections: Number of allowed concurrent connections, valid only if `asynchronous=True`.
+        :param total_timeout: Total timeout in secs for each of the concurrent requests when using `asynchronous=True`.
+        :param kwargs: Additional parameters to be sent along the request.
+        :return: List of HTTP POST responses
+        """
+
+        if body_batch:
+            assert (
+                query_batch is None
+            ), "'query_batch' has no effect if 'body_batch' is not None."
+        elif query_batch:
+            assert (
+                body_batch is None
+            ), "'body_batch' has no effect if 'query_batch' is not None."
+            assert (
+                query_model is not None
+            ), "Specify a 'query_model' when using 'query_batch' argument."
+            number_of_queries = len(query_batch)
+
+            if recall:
+                assert (
+                    len(recall) == number_of_queries
+                ), "Specify one recall tuple for each query in the batch."
+        else:
+            ValueError("Specify either 'query_batch' or 'body_batch'.")
+
+        if asynchronous:
+            coro = self._query_batch_async(
+                body_batch=body_batch,
+                query_batch=query_batch,
+                query_model=query_model,
+                recall=recall,
+                connections=connections,
+                total_timeout=total_timeout,
+                **kwargs,
+            )
+            return self._check_for_running_loop_and_run_coroutine(coro=coro)
+        else:
+            return self._query_batch_sync(
+                body_batch=body_batch,
+                query_batch=query_batch,
+                query_model=query_model,
+                recall=recall,
+                **kwargs,
             )
 
     def feed_data_point(self, schema: str, data_id: str, fields: Dict) -> VespaResponse:
@@ -243,8 +401,8 @@ class Vespa(object):
 
     def feed_batch(
         self,
-        schema: str,
         batch: List[Dict],
+        schema: Optional[str] = None,
         asynchronous=True,
         connections: Optional[int] = 100,
         total_timeout: int = 100,
@@ -252,13 +410,21 @@ class Vespa(object):
         """
         Feed a batch of data to a Vespa app.
 
-        :param schema: The schema that we are sending data to.
         :param batch: A list of dict containing the keys 'id' and 'fields' to be used in the :func:`feed_data_point`.
+        :param schema: The schema that we are sending data to. The schema is optional in case it is possible to infer
+            the schema from the application package.
         :param asynchronous: Set True to send data in async mode. Default to True.
         :param connections: Number of allowed concurrent connections, valid only if `asynchronous=True`.
         :param total_timeout: Total timeout in secs for each of the concurrent requests when using `asynchronous=True`.
         :return: List of HTTP POST responses
         """
+        if not schema:
+            try:
+                schema = self._infer_schema_name()
+            except ValueError:
+                raise ValueError(
+                    "Not possible to infer schema name. Specify schema parameter."
+                )
 
         if asynchronous:
             coro = self._feed_batch_async(
@@ -271,6 +437,18 @@ class Vespa(object):
         else:
             return self._feed_batch_sync(schema=schema, batch=batch)
 
+    def feed_df(self, df: DataFrame, include_id: bool = True, **kwargs):
+        """
+        Feed data contained in a DataFrame.
+
+        :param df: A DataFrame containing a required 'id' column and the remaining fields to be fed.
+        :param include_id: Include id on the fields to be fed. Default to True.
+        :param kwargs: Additional parameters are passed to :func:`feed_batch`.
+        :return: List of HTTP POST responses
+        """
+        batch = parse_feed_df(df=df, include_id=include_id)
+        return self.feed_batch(batch=batch, **kwargs)
+
     def delete_data(self, schema: str, data_id: str) -> VespaResponse:
         """
         Delete a data point from a Vespa app.
@@ -295,8 +473,8 @@ class Vespa(object):
 
     def delete_batch(
         self,
-        schema: str,
         batch: List[Dict],
+        schema: Optional[str] = None,
         asynchronous=True,
         connections: Optional[int] = 100,
         total_timeout: int = 100,
@@ -304,13 +482,22 @@ class Vespa(object):
         """
         Delete a batch of data from a Vespa app.
 
-        :param schema: The schema that we are deleting data from.
         :param batch: A list of dict containing the key 'id'.
+        :param schema: The schema that we are deleting data from. The schema is optional in case it is possible to infer
+            the schema from the application package.
         :param asynchronous: Set True to get data in async mode. Default to True.
         :param connections: Number of allowed concurrent connections, valid only if `asynchronous=True`.
         :param total_timeout: Total timeout in secs for each of the concurrent requests when using `asynchronous=True`.
         :return: List of HTTP POST responses
         """
+        if not schema:
+            try:
+                schema = self._infer_schema_name()
+            except ValueError:
+                raise ValueError(
+                    "Not possible to infer schema name. Specify schema parameter."
+                )
+
         if asynchronous:
             coro = self._delete_batch_async(
                 schema=schema,
@@ -359,8 +546,8 @@ class Vespa(object):
 
     def get_batch(
         self,
-        schema: str,
         batch: List[Dict],
+        schema: Optional[str] = None,
         asynchronous=True,
         connections: Optional[int] = 100,
         total_timeout: int = 100,
@@ -368,13 +555,22 @@ class Vespa(object):
         """
         Get a batch of data from a Vespa app.
 
-        :param schema: The schema that we are getting data from.
         :param batch: A list of dict containing the key 'id'.
+        :param schema: The schema that we are getting data from. The schema is optional in case it is possible to infer
+            the schema from the application package.
         :param asynchronous: Set True to get data in async mode. Default to True.
         :param connections: Number of allowed concurrent connections, valid only if `asynchronous=True`.
         :param total_timeout: Total timeout in secs for each of the concurrent requests when using `asynchronous=True`.
         :return: List of HTTP POST responses
         """
+        if not schema:
+            try:
+                schema = self._infer_schema_name()
+            except ValueError:
+                raise ValueError(
+                    "Not possible to infer schema name. Specify schema parameter."
+                )
+
         if asynchronous:
             coro = self._get_batch_async(
                 schema=schema,
@@ -424,8 +620,8 @@ class Vespa(object):
 
     def update_batch(
         self,
-        schema: str,
         batch: List[Dict],
+        schema: Optional[str] = None,
         asynchronous=True,
         connections: Optional[int] = 100,
         total_timeout: int = 100,
@@ -433,13 +629,22 @@ class Vespa(object):
         """
         Update a batch of data in a Vespa app.
 
-        :param schema: The schema that we are getting data from.
         :param batch: A list of dict containing the keys 'id', 'fields' and 'create' (create defaults to False).
+        :param schema: The schema that we are updating data to. The schema is optional in case it is possible to infer
+            the schema from the application package.
         :param asynchronous: Set True to update data in async mode. Default to True.
         :param connections: Number of allowed concurrent connections, valid only if `asynchronous=True`.
         :param total_timeout: Total timeout in secs for each of the concurrent requests when using `asynchronous=True`.
         :return: List of HTTP POST responses
         """
+        if not schema:
+            try:
+                schema = self._infer_schema_name()
+            except ValueError:
+                raise ValueError(
+                    "Not possible to infer schema name. Specify schema parameter."
+                )
+
         if asynchronous:
             coro = self._update_batch_async(
                 schema=schema,
@@ -488,7 +693,7 @@ class Vespa(object):
         fields: List[str],
         relevant_score: int = 1,
         default_score: int = 0,
-        **kwargs
+        **kwargs,
     ) -> List[Dict]:
         """
         Collect training data based on a single query
@@ -511,7 +716,7 @@ class Vespa(object):
             query=query,
             query_model=query_model,
             recall=(id_field, [relevant_id]),
-            **kwargs
+            **kwargs,
         )
         hits = relevant_id_result.hits
         features = []
@@ -521,7 +726,7 @@ class Vespa(object):
                     query=query,
                     query_model=query_model,
                     hits=number_additional_docs,
-                    **kwargs
+                    **kwargs,
                 )
                 hits.extend(random_hits_result.hits)
 
@@ -538,19 +743,43 @@ class Vespa(object):
 
     def collect_training_data(
         self,
-        labeled_data: List[Dict],
+        labeled_data: Union[List[Dict], DataFrame],
         id_field: str,
         query_model: QueryModel,
         number_additional_docs: int,
         relevant_score: int = 1,
         default_score: int = 0,
         show_progress: Optional[int] = None,
-        **kwargs
+        **kwargs,
     ) -> DataFrame:
         """
         Collect training data based on a set of labelled data.
 
-        :param labeled_data: Labelled data containing query, query_id and relevant ids.
+        labeled_data can be a DataFrame or a List of Dict:
+
+        >>> labeled_data_df = DataFrame(
+        ...     data={
+        ...         "qid": [0, 0, 1, 1],
+        ...         "query": ["Intrauterine virus infections and congenital heart disease", "Intrauterine virus infections and congenital heart disease", "Clinical and immunologic studies in identical twins discordant for systemic lupus erythematosus", "Clinical and immunologic studies in identical twins discordant for systemic lupus erythematosus"],
+        ...         "doc_id": [0, 3, 1, 5],
+        ...         "relevance": [1,1,1,1]
+        ...     }
+        ... )
+
+        >>> labeled_data = [
+        ...     {
+        ...         "query_id": 0,
+        ...         "query": "Intrauterine virus infections and congenital heart disease",
+        ...         "relevant_docs": [{"id": 0, "score": 1}, {"id": 3, "score": 1}]
+        ...     },
+        ...     {
+        ...         "query_id": 1,
+        ...         "query": "Clinical and immunologic studies in identical twins discordant for systemic lupus erythematosus",
+        ...         "relevant_docs": [{"id": 1, "score": 1}, {"id": 5, "score": 1}]
+        ...     }
+        ... ]
+
+        :param labeled_data: Labelled data containing query, query_id and relevant ids. See details about data format.
         :param id_field: The Vespa field representing the document id.
         :param query_model: Query model.
         :param number_additional_docs: Number of additional documents to retrieve for each relevant document.
@@ -563,6 +792,9 @@ class Vespa(object):
             and vespa rank features returned by the Query model RankProfile used.
         """
 
+        if isinstance(labeled_data, DataFrame):
+            labeled_data = parse_labeled_data(df=labeled_data)
+
         training_data = []
         number_queries = len(labeled_data)
         idx_total = 0
@@ -591,7 +823,7 @@ class Vespa(object):
                     number_additional_docs=number_additional_docs,
                     relevant_score=doc_data.get("score", relevant_score),
                     default_score=default_score,
-                    **kwargs
+                    **kwargs,
                 )
                 training_data.extend(training_data_point)
         training_data = DataFrame.from_records(training_data)
@@ -607,7 +839,7 @@ class Vespa(object):
         relevant_docs: List[Dict],
         default_score: int = 0,
         detailed_metrics=False,
-        **kwargs
+        **kwargs,
     ) -> Dict:
         """
         Evaluate a query according to evaluation metrics
@@ -648,7 +880,7 @@ class Vespa(object):
         detailed_metrics=False,
         per_query=False,
         aggregators=None,
-        **kwargs
+        **kwargs,
     ) -> DataFrame:
         """
         Evaluate a :class:`QueryModel` according to a list of :class:`EvalMetric`.
@@ -712,7 +944,7 @@ class Vespa(object):
                     relevant_docs=query_data["relevant_docs"],
                     default_score=default_score,
                     detailed_metrics=detailed_metrics,
-                    **kwargs
+                    **kwargs,
                 )
                 evaluation.append(evaluation_query)
         evaluation = DataFrame.from_records(evaluation)
@@ -735,27 +967,34 @@ class Vespa(object):
         else:
             return self._application_package
 
+    def get_default_query_model(self):
+        try:
+            app_package = self.application_package
+        except ValueError:
+            return None
+        return app_package.default_query_model
+
     def get_model_from_application_package(self, model_name: str):
         """Get model definition from application package, if available."""
         app_package = self.application_package
-        model = app_package.get_model(model_name=model_name)
+        model = app_package.get_model(model_id=model_name)
         return model
 
-    def predict(self, x, model_name, function_name="output_0"):
+    def predict(self, x, model_id, function_name="output_0"):
         """
         Obtain a stateless model evaluation.
 
         :param x: Input where the format depends on the task that the model is serving.
-        :param model_name: The name of the model used to serve the prediction.
+        :param model_id: The id of the model used to serve the prediction.
         :param function_name: The name of the output function to be evaluated.
         :return: Model prediction.
         """
-        model = self.get_model_from_application_package(model_name)
+        model = self.get_model_from_application_package(model_id)
         encoded_tokens = model.create_url_encoded_tokens(x=x)
         with VespaSync(self) as sync_app:
             return model.parse_vespa_prediction(
                 sync_app.predict(
-                    model_name=model_name,
+                    model_id=model_id,
                     function_name=function_name,
                     encoded_tokens=encoded_tokens,
                 )
@@ -765,6 +1004,10 @@ class Vespa(object):
 class VespaSync(object):
     def __init__(self, app: Vespa, pool_maxsize: int = 10) -> None:
         self.app = app
+        if self.app.key:
+            self.cert = (self.app.cert, self.app.key)
+        else:
+            self.cert = self.app.cert
         self.http_session = None
         self.adapter = HTTPAdapter(
             max_retries=retry_strategy, pool_maxsize=pool_maxsize
@@ -799,18 +1042,18 @@ class VespaSync(object):
         """
         end_point = "{}/ApplicationStatus".format(self.app.end_point)
         try:
-            response = self.http_session.get(end_point, cert=self.app.cert)
+            response = self.http_session.get(end_point, cert=self.cert)
         except ConnectionError:
             response = None
         return response
 
-    def get_model_endpoint(self, model_name: Optional[str] = None) -> Optional[dict]:
+    def get_model_endpoint(self, model_id: Optional[str] = None) -> Optional[dict]:
         """Get model evaluation endpoints."""
         end_point = "{}/model-evaluation/v1/".format(self.app.end_point)
-        if model_name:
-            end_point = end_point + model_name
+        if model_id:
+            end_point = end_point + model_id
         try:
-            response = self.http_session.get(end_point, cert=self.app.cert)
+            response = self.http_session.get(end_point, cert=self.cert)
             if response.status_code == 200:
                 return response.json()
             else:
@@ -819,20 +1062,20 @@ class VespaSync(object):
             response = None
         return response
 
-    def predict(self, model_name, function_name, encoded_tokens):
+    def predict(self, model_id, function_name, encoded_tokens):
         """
         Obtain a stateless model evaluation.
 
-        :param model_name: The name of the model used to serve the prediction.
+        :param model_id: The id of the model used to serve the prediction.
         :param function_name: The name of the output function to be evaluated.
         :param encoded_tokens: URL-encoded input to the model
         :return: Model prediction.
         """
         end_point = "{}/model-evaluation/v1/{}/{}/eval?{}".format(
-            self.app.end_point, model_name, function_name, encoded_tokens
+            self.app.end_point, model_id, function_name, encoded_tokens
         )
         try:
-            response = self.http_session.get(end_point, cert=self.app.cert)
+            response = self.http_session.get(end_point, cert=self.cert)
             if response.status_code == 200:
                 return response.json()
             else:
@@ -854,9 +1097,7 @@ class VespaSync(object):
             self.app.end_point, schema, schema, str(data_id)
         )
         vespa_format = {"fields": fields}
-        response = self.http_session.post(
-            end_point, json=vespa_format, cert=self.app.cert
-        )
+        response = self.http_session.post(end_point, json=vespa_format, cert=self.cert)
         return VespaResponse(
             json=response.json(),
             status_code=response.status_code,
@@ -871,7 +1112,7 @@ class VespaSync(object):
         query_model: Optional[QueryModel] = None,
         debug_request: bool = False,
         recall: Optional[Tuple] = None,
-        **kwargs
+        **kwargs,
     ) -> VespaQueryResponse:
         """
         Send a query request to the Vespa application.
@@ -898,7 +1139,7 @@ class VespaSync(object):
             )
         else:
             r = self.http_session.post(
-                self.app.search_end_point, json=body, cert=self.app.cert
+                self.app.search_end_point, json=body, cert=self.cert
             )
         return VespaQueryResponse(
             json=r.json(), status_code=r.status_code, url=str(r.url)
@@ -915,7 +1156,7 @@ class VespaSync(object):
         end_point = "{}/document/v1/{}/{}/docid/{}".format(
             self.app.end_point, schema, schema, str(data_id)
         )
-        response = self.http_session.delete(end_point, cert=self.app.cert)
+        response = self.http_session.delete(end_point, cert=self.cert)
         return VespaResponse(
             json=response.json(),
             status_code=response.status_code,
@@ -934,7 +1175,7 @@ class VespaSync(object):
         end_point = "{}/document/v1/{}/{}/docid/?cluster={}&selection=true".format(
             self.app.end_point, schema, schema, content_cluster_name
         )
-        response = self.http_session.delete(end_point, cert=self.app.cert)
+        response = self.http_session.delete(end_point, cert=self.cert)
         return response
 
     def get_data(self, schema: str, data_id: str) -> VespaResponse:
@@ -948,7 +1189,7 @@ class VespaSync(object):
         end_point = "{}/document/v1/{}/{}/docid/{}".format(
             self.app.end_point, schema, schema, str(data_id)
         )
-        response = self.http_session.get(end_point, cert=self.app.cert)
+        response = self.http_session.get(end_point, cert=self.cert)
         return VespaResponse(
             json=response.json(),
             status_code=response.status_code,
@@ -972,9 +1213,7 @@ class VespaSync(object):
             self.app.end_point, schema, schema, str(data_id), str(create).lower()
         )
         vespa_format = {"fields": {k: {"assign": v} for k, v in fields.items()}}
-        response = self.http_session.put(
-            end_point, json=vespa_format, cert=self.app.cert
-        )
+        response = self.http_session.put(end_point, json=vespa_format, cert=self.cert)
         return VespaResponse(
             json=response.json(),
             status_code=response.status_code,
@@ -1005,7 +1244,7 @@ class VespaAsync(object):
         sslcontext = False
         if self.app.cert is not None:
             sslcontext = ssl.create_default_context()
-            sslcontext.load_cert_chain(self.app.cert)
+            sslcontext.load_cert_chain(self.app.cert, self.app.key)
         conn = aiohttp.TCPConnector(ssl=sslcontext, limit=self.connections)
         self.aiohttp_session = aiohttp.ClientSession(
             connector=conn, timeout=aiohttp.ClientTimeout(total=self.total_timeout)
@@ -1017,11 +1256,12 @@ class VespaAsync(object):
             return
         return await self.aiohttp_session.close()
 
-    async def _wait(self, f, args):
-        tasks = [asyncio.create_task(f(*arg)) for arg in args]
+    async def _wait(self, f, args, **kwargs):
+        tasks = [asyncio.create_task(f(*arg, **kwargs)) for arg in args]
         await asyncio.wait(tasks, return_when=asyncio.ALL_COMPLETED)
         return [result for result in map(lambda task: task.result(), tasks)]
 
+    @retry(wait=wait_exponential(multiplier=1), stop=stop_after_attempt(3))
     async def query(
         self,
         body: Optional[Dict] = None,
@@ -1029,7 +1269,7 @@ class VespaAsync(object):
         query_model: Optional[QueryModel] = None,
         debug_request: bool = False,
         recall: Optional[Tuple] = None,
-        **kwargs
+        **kwargs,
     ):
         if debug_request:
             return self.app.query(
@@ -1045,6 +1285,52 @@ class VespaAsync(object):
             json=await r.json(), status_code=r.status, url=str(r.url)
         )
 
+    async def _query_semaphore(
+        self,
+        body: Optional[Dict],
+        query: Optional[str],
+        query_model: Optional[QueryModel],
+        recall: Optional[Tuple],
+        semaphore: asyncio.Semaphore,
+        **kwargs,
+    ):
+        async with semaphore:
+            return await self.query(
+                body=body, query=query, query_model=query_model, recall=recall, **kwargs
+            )
+
+    async def query_batch(
+        self,
+        body_batch: Optional[List[Dict]],
+        query_batch: Optional[List[str]],
+        query_model: Optional[QueryModel],
+        recall: Optional[List[Tuple]],
+        **kwargs,
+    ):
+        sem = asyncio.Semaphore(self.connections)
+        if body_batch:
+            return await self._wait(
+                self._query_semaphore,
+                [(body, None, None, None, sem) for body in body_batch],
+                **kwargs,
+            )
+        else:
+            if recall:
+                return await self._wait(
+                    self._query_semaphore,
+                    [
+                        (None, q, query_model, r, sem)
+                        for (q, r) in zip(query_batch, recall)
+                    ],
+                    **kwargs,
+                )
+            else:
+                return await self._wait(
+                    self._query_semaphore,
+                    [(None, q, query_model, None, sem) for q in query_batch],
+                    **kwargs,
+                )
+
     @retry(wait=wait_exponential(multiplier=1), stop=stop_after_attempt(3))
     async def feed_data_point(
         self, schema: str, data_id: str, fields: Dict
@@ -1079,6 +1365,7 @@ class VespaAsync(object):
             ],
         )
 
+    @retry(wait=wait_exponential(multiplier=1), stop=stop_after_attempt(3))
     async def delete_data(self, schema: str, data_id: str) -> VespaResponse:
         end_point = "{}/document/v1/{}/{}/docid/{}".format(
             self.app.end_point, schema, schema, str(data_id)
@@ -1104,6 +1391,7 @@ class VespaAsync(object):
             [(schema, data_point["id"], sem) for data_point in batch],
         )
 
+    @retry(wait=wait_exponential(multiplier=1), stop=stop_after_attempt(3))
     async def get_data(self, schema: str, data_id: str):
         end_point = "{}/document/v1/{}/{}/docid/{}".format(
             self.app.end_point, schema, schema, str(data_id)
@@ -1129,6 +1417,7 @@ class VespaAsync(object):
             [(schema, data_point["id"], sem) for data_point in batch],
         )
 
+    @retry(wait=wait_exponential(multiplier=1), stop=stop_after_attempt(3))
     async def update_data(
         self, schema: str, data_id: str, fields: Dict, create: bool = False
     ) -> VespaResponse:
