@@ -26,7 +26,7 @@ from vespa.package import ApplicationPackage, ModelServer
 class VespaDocker(ToJson, FromJson["VespaDocker"]):
     def __init__(
         self,
-        disk_folder: str,
+        disk_folder: Optional[str] = None,
         port: int = 8080,
         container_memory: Union[str, int] = 4 * (1024 ** 3),
         output_file: IO = sys.stdout,
@@ -35,7 +35,8 @@ class VespaDocker(ToJson, FromJson["VespaDocker"]):
         """
         Manage Docker deployments.
 
-        :param disk_folder: Disk folder to save the required Vespa config files.
+        :param disk_folder: Disk folder to save the required Vespa config files. Default to application name
+            folder within user's current working directory.
         :param port: Container port.
         :param output_file: Output file to write output messages.
         :param container_memory: Docker container memory available to the application.
@@ -135,6 +136,9 @@ class VespaDocker(ToJson, FromJson["VespaDocker"]):
         :param application_package: Application package to export.
         :return: None. Application package file will be stored on `disk_folder`.
         """
+        if not self.disk_folder:
+            self.disk_folder = os.path.join(os.getcwd(), application_package.name)
+
         Path(os.path.join(self.disk_folder, "application/schemas")).mkdir(
             parents=True, exist_ok=True
         )
@@ -257,6 +261,8 @@ class VespaDocker(ToJson, FromJson["VespaDocker"]):
         :param application_package: ApplicationPackage to be deployed.
         :return: a Vespa connection instance.
         """
+        if not self.disk_folder:
+            self.disk_folder = os.path.join(os.getcwd(), application_package.name)
 
         self.export_application_package(application_package=application_package)
 
@@ -281,6 +287,8 @@ class VespaDocker(ToJson, FromJson["VespaDocker"]):
             If None, we assume `disk_folder` to be the application folder.
         :return: a Vespa connection instance.
         """
+        if not self.disk_folder:
+            self.disk_folder = os.path.join(os.getcwd(), application_name)
 
         return self._execute_deployment(
             application_name=application_name,
@@ -694,15 +702,20 @@ class VespaCloud(object):
                 file=self.output,
             )
 
-    def deploy(self, instance: str, disk_folder: str) -> Vespa:
+    def deploy(self, instance: str, disk_folder: Optional[str] = None) -> Vespa:
         """
         Deploy the given application package as the given instance in the Vespa Cloud dev environment.
 
         :param instance: Name of this instance of the application, in the Vespa Cloud.
-        :param disk_folder: Disk folder to save the required Vespa config files.
+        :param disk_folder: Disk folder to save the required Vespa config files. Default to application name
+            folder within user's current working directory.
 
         :return: a Vespa connection instance.
         """
+
+        if not disk_folder:
+            disk_folder = os.path.join(os.getcwd(), self.application_package.name)
+
         region = self._get_dev_region()
         job = "dev-" + region
         run = self._start_deployment(instance, job, disk_folder)
