@@ -5,8 +5,10 @@ from unittest.mock import Mock, call
 from pandas import DataFrame
 from pandas.testing import assert_frame_equal
 
-from vespa.application import Vespa
-from vespa.query import QueryModel, OR, RankProfile, VespaResult
+from vespa.package import ApplicationPackage, ModelServer, Schema, Document
+from vespa.application import Vespa, parse_labeled_data, parse_feed_df
+from vespa.io import VespaQueryResponse
+from vespa.query import QueryModel, OR, RankProfile
 
 
 class TestVespa(unittest.TestCase):
@@ -21,6 +23,55 @@ class TestVespa(unittest.TestCase):
             Vespa(url="http://localhost/", port=8080).end_point, "http://localhost:8080"
         )
 
+    def test_infer_schema(self):
+        #
+        # No application package
+        #
+        app = Vespa(url="http://localhost", port=8080)
+        with self.assertRaisesRegex(
+            ValueError,
+            "Application Package not available. Not possible to infer schema name.",
+        ):
+            _ = app._infer_schema_name()
+        #
+        # No schema
+        #
+        app_package = ModelServer(name="test")
+        app = Vespa(url="http://localhost", port=8080, application_package=app_package)
+        with self.assertRaisesRegex(
+            ValueError,
+            "Application has no schema. Not possible to infer schema name.",
+        ):
+            _ = app._infer_schema_name()
+        #
+        # More than one schema
+        #
+        app_package = ApplicationPackage(
+            name="test",
+            schema=[
+                Schema(name="x", document=Document()),
+                Schema(name="y", document=Document()),
+            ],
+        )
+        app = Vespa(url="http://localhost", port=8080, application_package=app_package)
+        with self.assertRaisesRegex(
+            ValueError,
+            "Application has more than one schema. Not possible to infer schema name.",
+        ):
+            _ = app._infer_schema_name()
+        #
+        # One schema
+        #
+        app_package = ApplicationPackage(
+            name="test",
+            schema=[
+                Schema(name="x", document=Document()),
+            ],
+        )
+        app = Vespa(url="http://localhost", port=8080, application_package=app_package)
+        schema_name = app._infer_schema_name()
+        self.assertEqual("x", schema_name)
+
 
 class TestVespaQuery(unittest.TestCase):
     def test_query(self):
@@ -61,6 +112,136 @@ class TestVespaQuery(unittest.TestCase):
             },
         )
 
+    def test_query_with_body_function(self):
+        app = Vespa(url="http://localhost", port=8080)
+
+        def body_function(query):
+            body = {
+                "yql": "select * from sources * where userQuery();",
+                "query": query,
+                "type": "any",
+                "ranking": {"profile": "bm25", "listFeatures": "true"},
+            }
+            return body
+
+        query_model = QueryModel(body_function=body_function)
+
+        self.assertDictEqual(
+            app.query(
+                query="this is a test",
+                query_model=query_model,
+                debug_request=True,
+                hits=10,
+                recall=("id", [1, 5]),
+            ).request_body,
+            {
+                "yql": "select * from sources * where userQuery();",
+                "query": "this is a test",
+                "type": "any",
+                "ranking": {"profile": "bm25", "listFeatures": "true"},
+                "hits": 10,
+                "recall": "+(id:1 id:5)",
+            },
+        )
+
+
+class TestLabeledData(unittest.TestCase):
+    def test_parse_labeled_data(self):
+        labeled_data_df = DataFrame(
+            data={
+                "qid": [0, 0, 1, 1],
+                "query": ["Intrauterine virus infections and congenital heart disease"]
+                * 2
+                + [
+                    "Clinical and immunologic studies in identical twins discordant for systemic lupus erythematosus"
+                ]
+                * 2,
+                "doc_id": [0, 3, 1, 5],
+                "relevance": [1, 1, 1, 1],
+            }
+        )
+        labeled_data = parse_labeled_data(df=labeled_data_df)
+        expected_labeled_data = [
+            {
+                "query_id": 0,
+                "query": "Intrauterine virus infections and congenital heart disease",
+                "relevant_docs": [{"id": 0, "score": 1}, {"id": 3, "score": 1}],
+            },
+            {
+                "query_id": 1,
+                "query": "Clinical and immunologic studies in identical twins discordant for systemic lupus erythematosus",
+                "relevant_docs": [{"id": 1, "score": 1}, {"id": 5, "score": 1}],
+            },
+        ]
+        self.assertEqual(labeled_data, expected_labeled_data)
+
+    def test_parse_labeled_data_with_wrong_columns(self):
+        labeled_data_df = DataFrame(
+            data={
+                "qid": [0, 0, 1, 1],
+                "doc_id": [0, 3, 1, 5],
+                "relevance": [1, 1, 1, 1],
+            }
+        )
+        with self.assertRaises(AssertionError):
+            _ = parse_labeled_data(df=labeled_data_df)
+
+
+class TestParseFeedDataFrame(unittest.TestCase):
+    def setUp(self) -> None:
+        records = [
+            {
+                "id": idx,
+                "title": "This doc is about {}".format(x),
+                "body": "There is so much to learn about {}".format(x),
+            }
+            for idx, x in enumerate(
+                ["finance", "sports", "celebrity", "weather", "politics"]
+            )
+        ]
+        self.df = DataFrame.from_records(records)
+
+    def test_parse_simplified_feed_batch(self):
+        batch = parse_feed_df(self.df, include_id=True)
+        expected_batch = [
+            {
+                "id": idx,
+                "fields": {
+                    "id": idx,
+                    "title": "This doc is about {}".format(x),
+                    "body": "There is so much to learn about {}".format(x),
+                },
+            }
+            for idx, x in enumerate(
+                ["finance", "sports", "celebrity", "weather", "politics"]
+            )
+        ]
+        self.assertEqual(expected_batch, batch)
+
+    def test_parse_simplified_feed_batch_not_including_id(self):
+        batch = parse_feed_df(self.df, include_id=False)
+        expected_batch = [
+            {
+                "id": idx,
+                "fields": {
+                    "title": "This doc is about {}".format(x),
+                    "body": "There is so much to learn about {}".format(x),
+                },
+            }
+            for idx, x in enumerate(
+                ["finance", "sports", "celebrity", "weather", "politics"]
+            )
+        ]
+        self.assertEqual(expected_batch, batch)
+
+    def test_parse_simplified_feed_batch_with_wrong_columns(self):
+        missing_id_df = self.df[["title", "body"]]
+        with self.assertRaisesRegex(
+            AssertionError,
+            "DataFrame needs at least the following columns: \['id'\]",
+        ):
+            _ = parse_feed_df(df=missing_id_df, include_id=True)
+
 
 class TestVespaCollectData(unittest.TestCase):
     def setUp(self) -> None:
@@ -141,8 +322,12 @@ class TestVespaCollectData(unittest.TestCase):
 
         self.app.query = Mock(
             side_effect=[
-                VespaResult(self.raw_vespa_result_recall),
-                VespaResult(self.raw_vespa_result_additional),
+                VespaQueryResponse(
+                    self.raw_vespa_result_recall, status_code=None, url=None
+                ),
+                VespaQueryResponse(
+                    self.raw_vespa_result_additional, status_code=None, url=None
+                ),
             ]
         )
         query_model = QueryModel(rank_profile=RankProfile(list_features=True))
@@ -206,8 +391,12 @@ class TestVespaCollectData(unittest.TestCase):
 
         self.app.query = Mock(
             side_effect=[
-                VespaResult(self.raw_vespa_result_recall),
-                VespaResult(self.raw_vespa_result_additional),
+                VespaQueryResponse(
+                    self.raw_vespa_result_recall, status_code=None, url=None
+                ),
+                VespaQueryResponse(
+                    self.raw_vespa_result_additional, status_code=None, url=None
+                ),
             ]
         )
         query_model = QueryModel(rank_profile=RankProfile(list_features=True))
@@ -283,8 +472,12 @@ class TestVespaCollectData(unittest.TestCase):
         }
         self.app.query = Mock(
             side_effect=[
-                VespaResult(self.raw_vespa_result_recall),
-                VespaResult(self.raw_vespa_result_additional),
+                VespaQueryResponse(
+                    self.raw_vespa_result_recall, status_code=None, url=None
+                ),
+                VespaQueryResponse(
+                    self.raw_vespa_result_additional, status_code=None, url=None
+                ),
             ]
         )
         query_model = QueryModel(rank_profile=RankProfile(list_features=True))
@@ -452,24 +645,16 @@ class TestVespaEvaluate(unittest.TestCase):
         self.assertEqual(eval_metric.evaluate_query.call_count, 1)
         eval_metric.evaluate_query.assert_has_calls(
             [
-                call({}, self.labeled_data[0]["relevant_docs"], "vespa_id_field", 0),
+                call(
+                    {},
+                    self.labeled_data[0]["relevant_docs"],
+                    "vespa_id_field",
+                    0,
+                    False,
+                ),
             ]
         )
-        self.assertDictEqual(evaluation, {"query_id": "0", "metric": 1, "metric_2": 2})
-
-    def test_evaluate(self):
-        self.app.evaluate_query = Mock(
-            side_effect=[
-                {"query_id": "0", "metric": 1},
-            ]
-        )
-        evaluation = self.app.evaluate(
-            labeled_data=self.labeled_data,
-            eval_metrics=[Mock()],
-            query_model=Mock(),
-            id_field="mock",
-            default_score=0,
-        )
-        assert_frame_equal(
-            evaluation, DataFrame.from_records([{"query_id": "0", "metric": 1}])
+        self.assertDictEqual(
+            evaluation,
+            {"model": "default_name", "query_id": "0", "metric": 1, "metric_2": 2},
         )
