@@ -1,9 +1,11 @@
 import unittest
+import pytest
 import os
 import re
 import shutil
 import asyncio
 import json
+from pandas import DataFrame
 from vespa.package import (
     HNSW,
     Document,
@@ -18,7 +20,7 @@ from vespa.package import (
 from vespa.deployment import VespaDocker
 from vespa.ml import BertModelConfig, SequenceClassification
 from vespa.query import QueryModel, RankProfile as Ranking, OR, QueryRankingFeature
-from vespa.gallery import QuestionAnswering
+from vespa.gallery import QuestionAnswering, TextSearch
 from vespa.application import VespaSync
 
 
@@ -120,7 +122,7 @@ def create_qa_application_package():
 def create_sequence_classification_task():
     app_package = ModelServer(
         name="bert-model-server",
-        models=[
+        tasks=[
             SequenceClassification(
                 model_id="bert_tiny", model="google/bert_uncased_L-2_H-128_A-2"
             )
@@ -318,6 +320,12 @@ class TestApplicationCommon(unittest.TestCase):
                 ),
             },
         )
+        #
+        # Query with 'query' without QueryModel
+        #
+        with self.assertRaisesRegex(AssertionError, "No 'query_model' specified."):
+            _ = app.query(query="this should not work")
+
         #
         # Update data
         #
@@ -597,6 +605,10 @@ class TestApplicationCommon(unittest.TestCase):
         fields_to_send,
         expected_fields_from_get_operation,
         fields_to_update,
+        query_batch=None,
+        query_model=None,
+        hit_field_to_check=None,
+        queries_first_hit=None,
     ):
         """
         Sync feed a batch of data to the application
@@ -608,6 +620,10 @@ class TestApplicationCommon(unittest.TestCase):
         :param expected_fields_from_get_operation: Dict containing fields as returned by Vespa get operation.
             There are cases where fields returned from Vespa are different than inputs, e.g. when dealing with Tensors.
         :param fields_to_update: Dict where keys are field names and values are field values.
+        :param query_batch: Optional list of query strings.
+        :param query_model: Optional QueryModel to use with query_batch.
+        :param hit_field_to_check: Which field of the query response should be checked.
+        :param queries_first_hit: The expected field of the first hit of each query sent
         :return:
         """
 
@@ -631,6 +647,18 @@ class TestApplicationCommon(unittest.TestCase):
         )
         self.assertEqual(result.number_documents_indexed, num_docs)
 
+        #
+        # Query data
+        #
+        if query_batch:
+            result = app.query_batch(
+                query_batch=query_batch, query_model=query_model, asynchronous=False
+            )
+            for idx, first_hit in enumerate(queries_first_hit):
+                self.assertEqual(
+                    first_hit, result[idx].hits[0]["fields"][hit_field_to_check]
+                )
+
         #
         # get batch data
         #
@@ -685,6 +713,10 @@ class TestApplicationCommon(unittest.TestCase):
         fields_to_send,
         expected_fields_from_get_operation,
         fields_to_update,
+        query_batch=None,
+        query_model=None,
+        hit_field_to_check=None,
+        queries_first_hit=None,
     ):
         """
         Async feed a batch of data to the application
@@ -696,6 +728,10 @@ class TestApplicationCommon(unittest.TestCase):
         :param expected_fields_from_get_operation: Dict containing fields as returned by Vespa get operation.
             There are cases where fields returned from Vespa are different than inputs, e.g. when dealing with Tensors.
         :param fields_to_update: Dict where keys are field names and values are field values.
+        :param query_batch: Optional list of query strings.
+        :param query_model: Optional QueryModel to use with query_batch.
+        :param hit_field_to_check: Which field of the query response should be checked.
+        :param queries_first_hit: The expected field of the first hit of each query sent
         :return:
         """
         #
@@ -724,6 +760,18 @@ class TestApplicationCommon(unittest.TestCase):
         )
         self.assertEqual(result.number_documents_indexed, num_docs)
 
+        #
+        # Query data
+        #
+        if query_batch:
+            result = app.query_batch(
+                query_batch=query_batch, query_model=query_model
+            )
+            for idx, first_hit in enumerate(queries_first_hit):
+                self.assertEqual(
+                    first_hit, result[idx].hits[0]["fields"][hit_field_to_check]
+                )
+
         #
         # get batch data
         #
@@ -771,6 +819,93 @@ class TestApplicationCommon(unittest.TestCase):
         for idx, response in enumerate(result):
             self.assertEqual(response.status_code, 404)
 
+    def batch_operations_default_mode_with_one_schema(
+        self,
+        app,
+        schema_name,
+        fields_to_send,
+        expected_fields_from_get_operation,
+        fields_to_update,
+    ):
+        """
+        Document batch operations for applications with one schema
+
+        :param app: Vespa instance holding the connection to the application
+        :param schema_name: Schema name containing the document we want to send and retrieve data
+        :param fields_to_send: List of Dicts where keys are field names and values are field values. Must
+            contain 'id' field.
+        :param expected_fields_from_get_operation: Dict containing fields as returned by Vespa get operation.
+            There are cases where fields returned from Vespa are different than inputs, e.g. when dealing with Tensors.
+        :param fields_to_update: Dict where keys are field names and values are field values.
+        :return:
+        """
+        #
+        # Create and feed documents
+        #
+        num_docs = len(fields_to_send)
+        schema = schema_name
+        docs = [{"id": fields["id"], "fields": fields} for fields in fields_to_send]
+        update_docs = [
+            {"id": fields["id"], "fields": fields} for fields in fields_to_update
+        ]
+
+        app.feed_batch(batch=docs)
+
+        #
+        # Verify that all documents are fed
+        #
+        result = app.query(
+            query="sddocname:{}".format(schema_name), query_model=QueryModel()
+        )
+        self.assertEqual(result.number_documents_indexed, num_docs)
+
+        #
+        # get batch data
+        #
+        result = app.get_batch(batch=docs)
+        for idx, response in enumerate(result):
+            self.assertDictEqual(
+                response.json["fields"], expected_fields_from_get_operation[idx]
+            )
+
+        #
+        # Update data
+        #
+        result = app.update_batch(batch=update_docs)
+        for idx, response in enumerate(result):
+            self.assertEqual(
+                response.json["id"],
+                "id:{}:{}::{}".format(schema, schema, fields_to_update[idx]["id"]),
+            )
+
+        #
+        # Get updated data
+        #
+        result = app.get_batch(batch=docs)
+        for idx, response in enumerate(result):
+            expected_updated_fields = {
+                k: v for k, v in expected_fields_from_get_operation[idx].items()
+            }
+            expected_updated_fields.update(fields_to_update[idx])
+            self.assertDictEqual(response.json["fields"], expected_updated_fields)
+
+        #
+        # Delete data
+        #
+        result = app.delete_batch(batch=docs)
+        for idx, response in enumerate(result):
+            self.assertEqual(
+                response.json["id"],
+                "id:{}:{}::{}".format(schema, schema, docs[idx]["id"]),
+            )
+
+        #
+        # get batch deleted data
+        #
+        result = app.get_batch(batch=docs)
+        for idx, response in enumerate(result):
+            self.assertEqual(response.status_code, 404)
+
     def get_model_endpoints_when_no_model_is_available(
         self, app, expected_model_endpoint
     ):
@@ -782,7 +917,7 @@ class TestApplicationCommon(unittest.TestCase):
             },
         )
         self.assertEqual(
-            app.get_model_endpoint(model_name="bert_tiny"),
+            app.get_model_endpoint(model_id="bert_tiny"),
             {
                 "status_code": 404,
                 "message": "No binding for URI '{}bert_tiny'.".format(
@@ -797,11 +932,11 @@ class TestApplicationCommon(unittest.TestCase):
             {"bert_tiny": "{}bert_tiny".format(expected_model_endpoint)},
         )
         self.assertEqual(
-            app.get_model_endpoint(model_name="bert_tiny")["model"], "bert_tiny"
+            app.get_model_endpoint(model_id="bert_tiny")["model"], "bert_tiny"
         )
 
     def get_stateless_prediction(self, app, application_package):
-        prediction = app.predict("this is a test", model_name="bert_tiny")
+        prediction = app.predict("this is a test", model_id="bert_tiny")
         expected_values = application_package.models["bert_tiny"].predict(
             "this is a test"
         )
@@ -812,7 +947,7 @@ class TestApplicationCommon(unittest.TestCase):
         with self.assertRaisesRegex(
             ValueError, "Model named bert_tiny not defined in the application package"
         ) as exc:
-            _ = app.predict("this is a test", model_name="bert_tiny")
+            _ = app.predict("this is a test", model_id="bert_tiny")
 
     @staticmethod
     def _parse_vespa_tensor(hit, feature):
@@ -904,6 +1039,7 @@ class TestMsmarcoDockerDeployment(TestDockerCommon):
             application_package=self.app_package, disk_folder=self.disk_folder
         )
 
+    @pytest.mark.skip(reason="Works locally but fails on Screwdriver")
     def test_redeploy_with_container_stopped(self):
         self.redeploy_with_container_stopped(
             application_package=self.app_package, disk_folder=self.disk_folder
@@ -974,6 +1110,11 @@ class TestMsmarcoApplication(TestApplicationCommon):
             }
             for i in range(10)
         ]
+        self.query_batch = ["Give me title 1", "Give me title 2"]
+        self.query_model = QueryModel(
+            match_phase=OR(), rank_profile=Ranking(name="default", list_features=False)
+        )
+        self.queries_first_hit = ["this is title 1", "this is title 2"]
 
     def test_model_endpoints_when_no_model_is_available(self):
         # The port should be 8089 instead of 8080, see https://jira.vzbuilders.com/browse/VESPA-21365
@@ -1014,6 +1155,10 @@ class TestMsmarcoApplication(TestApplicationCommon):
             fields_to_send=self.fields_to_send,
             expected_fields_from_get_operation=self.fields_to_send,
             fields_to_update=self.fields_to_update,
+            query_batch=self.query_batch,
+            query_model=self.query_model,
+            hit_field_to_check="title",
+            queries_first_hit=self.queries_first_hit,
         )
 
     def test_batch_operations_asynchronous_mode(self):
@@ -1023,6 +1168,19 @@ class TestMsmarcoApplication(TestApplicationCommon):
             fields_to_send=self.fields_to_send,
             expected_fields_from_get_operation=self.fields_to_send,
             fields_to_update=self.fields_to_update,
+            query_batch=self.query_batch,
+            query_model=self.query_model,
+            hit_field_to_check="title",
+            queries_first_hit=self.queries_first_hit,
+        )
+
+    def test_batch_operations_default_mode_with_one_schema(self):
+        self.batch_operations_default_mode_with_one_schema(
+            app=self.app,
+            schema_name=self.app_package.name,
+            fields_to_send=self.fields_to_send,
+            expected_fields_from_get_operation=self.fields_to_send,
+            fields_to_update=self.fields_to_update,
         )
 
     def tearDown(self) -> None:
@@ -1130,6 +1288,15 @@ class TestCord19Application(TestApplicationCommon):
             fields_to_update=self.fields_to_update,
         )
 
+    def test_batch_operations_default_mode_with_one_schema(self):
+        self.batch_operations_default_mode_with_one_schema(
+            app=self.app,
+            schema_name=self.app_package.name,
+            fields_to_send=self.fields_to_send,
+            expected_fields_from_get_operation=self.expected_fields_from_get_operation,
+            fields_to_update=self.fields_to_update,
+        )
+
     def test_bert_model_input_and_output(self):
         self.bert_model_input_and_output(
             app=self.app,
@@ -1257,6 +1424,56 @@ class TestQaApplication(TestApplicationCommon):
         self.vespa_docker.container.remove()
 
 
+class TestGalleryTextSearch(unittest.TestCase):
+    def setUp(self) -> None:
+        #
+        # Create application
+        #
+        self.app_package = TextSearch(id_field="id", text_fields=["title", "body"])
+        #
+        # Deploy application
+        #
+        self.disk_folder = os.path.join(os.getenv("WORK_DIR"), "sample_application")
+        self.vespa_docker = VespaDocker(port=8089, disk_folder=self.disk_folder)
+        self.app = self.vespa_docker.deploy(application_package=self.app_package)
+        #
+        # Create a sample data frame
+        #
+        records = [
+            {
+                "id": idx,
+                "title": "This doc is about {}".format(x),
+                "body": "There is so much to learn about {}".format(x),
+            }
+            for idx, x in enumerate(
+                ["finance", "sports", "celebrity", "weather", "politics"]
+            )
+        ]
+        df = DataFrame.from_records(records)
+        #
+        # Feed application
+        #
+        self.app.feed_df(df)
+
+    def test_default_query_model(self):
+        result = self.app.query(query="what is finance?", debug_request=True)
+        expected_request_body = {
+            "yql": 'select * from sources * where (userInput("what is finance?"));',
+            "ranking": {"profile": "bm25", "listFeatures": "false"},
+        }
+        self.assertDictEqual(expected_request_body, result.request_body)
+
+    def test_query(self):
+        result = self.app.query(query="what is finance?")
+        for hit in result.hits:
+            self.assertIn("fields", hit)
+
+    def tearDown(self) -> None:
+        shutil.rmtree(self.disk_folder, ignore_errors=True)
+        self.vespa_docker.container.stop()
+        self.vespa_docker.container.remove()
+
+
 class TestSequenceClassification(TestApplicationCommon):
     def setUp(self) -> None:
         self.app_package = create_sequence_classification_task()
